using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using UnityEngine.Events;

namespace Stopper {
	public class StaticClothing : MVRScript {

        string fileExt = "staticclothing";

        DAZClothingItem dazClothingItem;
        DAZDynamic dazDynamic;
        DAZMesh dazMesh;
        DAZSkinWrap dazSkinWrap;

        MeshDrawOffsetControl meshDrawOffsetControl;
        DAZMeshDrawOffsetControlUI meshDrawOffsetControlUI;

        DAZSkinWrapMaterialOptions materialOptions;

        DAZBone drawFromBone;

        DummySkinWrap dummySkinWrap;
        Dictionary<DAZSkinWrapMaterialOptions, DAZSkinWrap> originalMatOptsSkin = new Dictionary<DAZSkinWrapMaterialOptions, DAZSkinWrap>();

        Material[] dazMeshOriginalMaterials;

        JSONStorableFloat baseOffsetX;        
        JSONStorableFloat baseOffsetY;
        JSONStorableFloat baseOffsetZ;

        bool needsUpdate = false;

        string overrideStoreId;

        public class MeshDrawOffsetControl : DAZMeshDrawOffsetControl {
            public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true){
                return;
            }
        }

        public class DummySkinWrap : DAZSkinWrap {
            public DummySkinWrap(){
                _materialsWereInit = true;
            }

            public void SetMaterials(Material[] materials) {
                this.GPUmaterials = materials;
                this._numMaterials = materials.Length;
            }
        }
        
		public override void Init() {
			WalkAndGetParentItems();

            if(dazClothingItem == null) {
                SuperController.LogError($"Could not find skinwrap. Is this a clothing item?");
                return;
            }

            if(dazSkinWrap == null) {
                SuperController.LogError($"Could not find skinwrap. Is this a clothing item?");
                return;
            }

            dazMeshOriginalMaterials = new Material[dazSkinWrap.dazMesh.materials.Length];
                
            for(int i = 0; i < dazSkinWrap.dazMesh.materials.Length; i++){
                dazMeshOriginalMaterials[i] = dazSkinWrap.dazMesh.materials[i];
            }

            this.overrideId = $"{dazSkinWrap.name}:{this.storeId}";

            drawFromBone = containingAtom.GetStorableByID("head") as DAZBone;

            meshDrawOffsetControl = gameObject.AddComponent<MeshDrawOffsetControl>();
            meshDrawOffsetControl.mesh = dazSkinWrap.dazMesh;
            
            meshDrawOffsetControlUI = gameObject.AddComponent<DAZMeshDrawOffsetControlUI>();

            // Initializing this via constructor since we don't want it to interact with anything.
            dummySkinWrap = gameObject.AddComponent<DummySkinWrap>();
            
            foreach(var materialOptions in dazSkinWrap.GetComponents<DAZSkinWrapMaterialOptions>()) {
                originalMatOptsSkin[materialOptions] = materialOptions.skinWrap;
            }

            // Preset code was adapted from AcidBubbles' Collider Editor
            var savePath = dazDynamic.GetStoreFolderPath(false);
			var loadPresetUI = CreateButton("Load Preset");
			loadPresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var shortcuts = FileManagerSecure.GetShortCutsForDirectory(savePath);
				SuperController.singleton.GetMediaPathDialog(LoadPreset, fileExt, savePath, false, true, false, null, false, shortcuts);
			});

			var savePresetUI = CreateButton("Save Preset");
			savePresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var fileBrowserUI = SuperController.singleton.fileBrowserUI;
				fileBrowserUI.SetTitle("Save Contortionlist Preset");
				fileBrowserUI.fileRemovePrefix = null;
				fileBrowserUI.hideExtension = false;
				fileBrowserUI.keepOpen = false;
				fileBrowserUI.fileFormat = fileExt;
				fileBrowserUI.defaultPath = savePath;
				fileBrowserUI.showDirs = true;
				fileBrowserUI.shortCuts = null;
				fileBrowserUI.browseVarFilesAsDirectories = false;
				fileBrowserUI.SetTextEntry(true);
				fileBrowserUI.Show(SavePreset);
				fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + fileExt;
				fileBrowserUI.ActivateFileNameField();
			});
            
            // Origin
            

            // Position
            var xPositionSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("positionOffset:x"),null);
            meshDrawOffsetControlUI.xPositionOffsetSlider = xPositionSlider.slider;

            var yPositionSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("positionOffset:y"),null);
            meshDrawOffsetControlUI.yPositionOffsetSlider = yPositionSlider.slider;

            var zPositionSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("positionOffset:z"),null);
            meshDrawOffsetControlUI.zPositionOffsetSlider = zPositionSlider.slider;

            // Scale
            var xScaleSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("scale:x"),null);
            meshDrawOffsetControlUI.xScaleSlider = xScaleSlider.slider;

            var yScaleSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("scale:y"),null);
            meshDrawOffsetControlUI.yScaleSlider = yScaleSlider.slider;

            var zScaleSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("scale:z"),null);
            meshDrawOffsetControlUI.zScaleSlider = zScaleSlider.slider;

            var overallScaleSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("scale:overall"),null);
            meshDrawOffsetControlUI.overallScaleSlider = overallScaleSlider.slider;


            // Bone
            List<string> boneNames = new List<string>();
            foreach(var storableId in containingAtom.GetStorableIDs()){
                var storable = containingAtom.GetStorableByID(storableId);
                if(storable is DAZBone){
                    boneNames.Add(storableId);
                }
            }

            JSONStorableStringChooser parentBoneChooser = new JSONStorableStringChooser(
            "Parent Bone", 
            boneNames,
            "head", 
            "Parent Bone", 
            (string boneName) => {
                drawFromBone = containingAtom.GetStorableByID(boneName) as DAZBone;
                dazSkinWrap.dazMesh.drawFromBone = drawFromBone;
            });
			CreateScrollablePopup(parentBoneChooser, true);
            RegisterStringChooser(parentBoneChooser);

            // Rotation
            baseOffsetX = new JSONStorableFloat("Rotation Center Offset X", 0f,-1f,1f, false);
            var xBaseOffsetSlider = AddSlider(baseOffsetX,UpdateDrawOffsetOrigin,true,true);

            baseOffsetY = new JSONStorableFloat("Rotation Center Offset Y", 0f,-1f,1f, false);
            var yBaseOffsetSlider = AddSlider(baseOffsetY,UpdateDrawOffsetOrigin,true,true);

            baseOffsetZ = new JSONStorableFloat("Rotation Center Offset Z", 0f,-1f,1f, false);
            var zBaseOffsetSlider = AddSlider(baseOffsetZ,UpdateDrawOffsetOrigin,true,true);

            var xRotationSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("rotationOffset:x"),null,true,true);
            meshDrawOffsetControlUI.xRotationOffsetSlider = xRotationSlider.slider;

            var yRotationSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("rotationOffset:y"),null,true,true);
            meshDrawOffsetControlUI.yRotationOffsetSlider = yRotationSlider.slider;

            var zRotationSlider = AddSlider(meshDrawOffsetControl.GetFloatJSONParam("rotationOffset:z"),null,true,true);
            meshDrawOffsetControlUI.zRotationOffsetSlider = zRotationSlider.slider;

            meshDrawOffsetControl.UITransform = this.transform;
            meshDrawOffsetControl.InitUI();

            OnEnable();

            // Try to load the default preset
            try {
                string defaultFile = $"{dazDynamic.GetStoreFolderPath(true)}default.{fileExt}";
                if(FileManagerSecure.FileExists(defaultFile)){
                    LoadPreset(defaultFile);
                } 
            } catch (Exception e) {
                SuperController.LogMessage($"Error loading default {e}");
                // eat
            }
		}
		
        private void WalkAndGetParentItems(){
			GameObject parent = this.gameObject;
			while(containingAtom == null || parent != containingAtom.gameObject){
				if(parent.transform.parent == null){
					return;
				}
				
				if(dazClothingItem == null){
					dazClothingItem = parent.GetComponent<DAZClothingItem>();
				}

                if(dazMesh == null){
					dazMesh = parent.GetComponent<DAZMesh>();
				}

                if(dazDynamic == null){
					dazDynamic = parent.GetComponent<DAZDynamic>();
				}

				if(dazSkinWrap == null){
					dazSkinWrap = parent.GetComponent<DAZSkinWrap>();
				}

                containingAtom = parent.GetComponent<Atom>();
				parent = parent.transform.parent.gameObject;
			}
		}

        public UIDynamicSlider AddSlider(JSONStorableFloat storableFloat, UnityAction<float> onChange = null, bool rangeAdjust = true, bool rightSide = false){
            UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
            uiSlider.rangeAdjustEnabled = rangeAdjust;
			uiSlider.valueFormat = "F4";
            RegisterFloat(storableFloat);
            
            if(onChange != null){
				uiSlider.slider.onValueChanged.AddListener(onChange);
			}
			
			return uiSlider;
		}

        private void UpdateDrawOffsetOrigin(float val) {
            if(dazSkinWrap != null && drawFromBone != null){
                dazSkinWrap.dazMesh.drawOffsetOrigin = new Vector3(baseOffsetX.val, baseOffsetY.val, baseOffsetZ.val);
            }
        }
        
		private void OnEnable() {
            if(dazSkinWrap != null && drawFromBone != null){
                dazSkinWrap.enabled = false;
                dazSkinWrap.dazMesh.useDrawOffset = true;
                dazSkinWrap.dazMesh.enabled = true;
                dazSkinWrap.dazMesh.drawMorphedUVMappedMesh = true;
                dazSkinWrap.dazMesh.delayDisplayFrameCount = 1;
                dazSkinWrap.dazMesh.drawFromBone = drawFromBone;
                dazSkinWrap.dazMesh.drawOffsetOrigin = new Vector3(baseOffsetX.val, baseOffsetY.val, baseOffsetZ.val);
                
                for(int i = 0; i < dazSkinWrap.dazMesh.materials.Length; i++){
                    dazSkinWrap.dazMesh.materials[i] = new Material(dazSkinWrap.GPUmaterials[i]);
                    dazSkinWrap.dazMesh.materials[i].shader = Shader.Find("Custom/Subsurface/TransparentGlossNMNoCullSeparateAlpha");
                }

                foreach(var materialOptions in dazSkinWrap.GetComponents<DAZSkinWrapMaterialOptions>()) {
                    dummySkinWrap.GPUsimpleMaterial = dazSkinWrap.dazMesh.materials[0];
                    dummySkinWrap.SetMaterials(dazSkinWrap.dazMesh.materials);
                    materialOptions.skinWrap = dummySkinWrap;
                }
                
            }
        }
		private void OnDisable() {
            if(dazSkinWrap != null){
                dazSkinWrap.enabled = true;
                dazSkinWrap.dazMesh.enabled = false;
                dazSkinWrap.dazMesh.drawMorphedUVMappedMesh = false;

                if(dazMeshOriginalMaterials != null) {
                    for(int i = 0; i < dazSkinWrap.dazMesh.materials.Length; i++){
                        if(dazMeshOriginalMaterials[i] != null) {
                            dazSkinWrap.dazMesh.materials[i] = dazMeshOriginalMaterials[i];
                        }
                    }
                }

                foreach(var entry in originalMatOptsSkin) {
                    entry.Key.skinWrap = entry.Value;
                }
            }
		}

        private void OnDestroy() {
            if(meshDrawOffsetControl != null) Destroy(meshDrawOffsetControl);
            if(meshDrawOffsetControlUI != null) Destroy(meshDrawOffsetControlUI);
        }

        /* 
            I can't find a way to get the settings to apply that's better than this..
        */
        private IEnumerator Nudge() {
            yield return null;
            while(meshDrawOffsetControl == null){
                yield return null;
            }
            meshDrawOffsetControl.overallScale += 0.001f;
            yield return null;
            meshDrawOffsetControl.overallScale -= 0.001f;
            yield break;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true){
            base.RestoreFromJSON(jc,restorePhysical,restoreAppearance,presetAtoms,false);
            
            try{
                StartCoroutine("Nudge");
            } catch (Exception e){
                // eat
            }
        }

        private void LoadPreset(string path) {
			if (string.IsNullOrEmpty(path))
				return;
			
			var presetJson = (JSONClass)LoadJSON(path);
			
			RestoreFromJSON(presetJson, true, true, null, false);
			
			
		}
		
		private void SavePreset(string path) {
			SuperController.singleton.fileBrowserUI.fileFormat = null;
			if (string.IsNullOrEmpty(path)) {
				return;
			}

			if (!path.ToLower().EndsWith($".{fileExt}")) {
				path += $".{fileExt}";
			}

			SaveJSON(GetJSON(true,true,true), path);
		}
	}
}